package org.raku.nqp.sixmodel;

public final class ParameterizedType extends AbstractParametricity {
    /* The type that we are a parameterization of. */
    public SixModelObject parametricType;

    /* Our type parameters. */
    public SixModelObject parameters;
}
