module.exports = {
    "extends": "google",
    "parserOptions": {
      "ecmaVersion": 10
    },
    "rules": {
      "require-jsdoc": "off",
      "max-len": "off",
      "arrow-parens": "off",
      "camelcase": "off",
      "guard-for-in": "off",
      "prefer-rest-params": "off",
      "prefer-spread": "off",
      "no-invalid-this": "off",
      "no-throw-literal": "off",
      "prefer-const": "error",
      "spaced-comment": "off",
      "new-cap": "off"
    }
};
