plan(35);
class Reserved {
    method break() {'break'}
    method case() {'case'}
    method catch() {'catch'}
    method continue() {'continue'}
    method debugger() {'debugger'}
    method default() {'default'}
    method delete() {'delete'}
    method do() {'do'}
    method else() {'else'}
    method finally() {'finally'}
    method for() {'for'}
    method function() {'function'}
    method if() {'if'}
    method in() {'in'}
    method instanceof() {'instanceof'}
    method new() {'new'}
    method return() {'return'}
    method switch() {'switch'}
    method this() {'this'}
    method throw() {'throw'}
    method try() {'try'}
    method typeof() {'typeof'}
    method var() {'var'}
    method void() {'void'}
    method while() {'while'}
    method with() {'with'}
    method implements() {'implements'}
    method interface() {'interface'}
    method let() {'let'}
    method package() {'package'}
    method private() {'private'}
    method protected() {'protected'}
    method public() {'public'}
    method static() {'static'}
    method yield() {'yield'}
}
is(Reserved.break, 'break');
is(Reserved.case, 'case');
is(Reserved.catch, 'catch');
is(Reserved.continue, 'continue');
is(Reserved.debugger, 'debugger');
is(Reserved.default, 'default');
is(Reserved.delete, 'delete');
is(Reserved.do, 'do');
is(Reserved.else, 'else');
is(Reserved.finally, 'finally');
is(Reserved.for, 'for');
is(Reserved.function, 'function');
is(Reserved.if, 'if');
is(Reserved.in, 'in');
is(Reserved.instanceof, 'instanceof');
is(Reserved.new, 'new');
is(Reserved.return, 'return');
is(Reserved.switch, 'switch');
is(Reserved.this, 'this');
is(Reserved.throw, 'throw');
is(Reserved.try, 'try');
is(Reserved.typeof, 'typeof');
is(Reserved.var, 'var');
is(Reserved.void, 'void');
is(Reserved.while, 'while');
is(Reserved.with, 'with');
is(Reserved.implements, 'implements');
is(Reserved.interface, 'interface');
is(Reserved.let, 'let');
is(Reserved.package, 'package');
is(Reserved.private, 'private');
is(Reserved.protected, 'protected');
is(Reserved.public, 'public');
is(Reserved.static, 'static');
is(Reserved.yield, 'yield');
