package org.raku.nqp.sixmodel.reprs;

import org.raku.nqp.sixmodel.SixModelObject;

public class KnowHOWAttributeInstance extends SixModelObject {
    public String name;
    public SixModelObject type;
    public int box_target;
}
